@@ -18,9 +18,6 @@
 /* Define as 1 if you have catgets and don't want to use GNU gettext.  */
 #undef HAVE_CATGETS
 
-/* Define if you have GNU's getopt family of functions.  */
-#undef HAVE_GETOPT_LONG
-
 /* Define as 1 if you have gettext and don't want to use GNU gettext.  */
 #undef HAVE_GETTEXT
 
