@@ -15,6 +15,7 @@
 #endif /* HAVE_MEMORY_H */
 #if defined(STDC_HEADERS)
 #include  <stdlib.h>
+#include  <limits.h>
 #endif
 #if defined(HAVE_UNISTD_H)
 #include <unistd.h>
@@ -24,6 +25,9 @@
 #endif
 #include  <sys/time.h>
 #include  <signal.h>
+#ifdef HAVE_SYS_WAIT_H
+#include <sys/wait.h>
+#endif
 
 #ifdef HAVE_NET_SOCKET_H
 #include <net/socket.h>
@@ -63,6 +67,7 @@ int stage;		/* where are we? */
 int phase;		/* where are we, for error-logging purposes? */
 int batchcount;		/* count of messages sent in current batch */
 flag peek_capable;	/* can we peek for better error recovery? */
+int mailserver_socket_temp = -1;	/* socket to free if connect timeout */ 
 
 static int timeoutcount;		/* count consecutive timeouts */
 
@@ -437,9 +442,7 @@ static int fetch_messages(int mailserver_socket, struct query *ctl,
 		    report_build(stdout, GT_(" (length -1)"));
 		    break;
 		case MSGLEN_TOOLARGE:
-		    report_build(stdout, 
-				 GT_(" (oversized, %d octets)"),
-				 msgsizes[num-1]);
+		    report_build(stdout, GT_(" (oversized)"));
 		    break;
 		}
 	    }
@@ -452,8 +455,8 @@ static int fetch_messages(int mailserver_socket, struct query *ctl,
 	    err = (ctl->server.base_protocol->fetch_headers)(mailserver_socket,ctl,num, &len);
 	    if (err == PS_TRANSIENT)    /* server is probably Exchange */
 	    {
-		report_build(stdout,
-			     GT_("couldn't fetch headers, message %s@%s:%d (%d octets)"),
+		report(stdout,
+			     GT_("couldn't fetch headers, message %s@%s:%d (%d octets)\n"),
 			     ctl->remotename, ctl->server.truename, num,
 			     msgsizes[num-1]);
 		continue;
@@ -490,15 +493,28 @@ static int fetch_messages(int mailserver_socket, struct query *ctl,
 	    err = readheaders(mailserver_socket, len, msgsizes[num-1],
 			     ctl, num);
 	    if (err == PS_RETAINED)
-		suppress_forward = retained = TRUE;
+		suppress_readbody = suppress_forward = suppress_delete = retained = TRUE;
 	    else if (err == PS_TRANSIENT)
 		suppress_delete = suppress_forward = TRUE;
 	    else if (err == PS_REFUSED)
 		suppress_forward = TRUE;
+#if 0
+	    /* 
+	     * readheaders does not read the body when it
+	     * hits a non-header. It has been recently
+	     * fixed to return PS_TRUNCATED (properly) when
+	     * that happens, but apparently fixing that bug
+	     * opened this one here (which looks like an 
+	     * inproper fix from some ancient thinko)
+	     */
 	    else if (err == PS_TRUNCATED)
 		suppress_readbody = TRUE;
 	    else if (err)
 		return(err);
+#else
+	    else if (err && err != PS_TRUNCATED)
+		return(err);
+#endif
 
 	    /* 
 	     * If we're using IMAP4 or something else that
@@ -543,11 +559,6 @@ static int fetch_messages(int mailserver_socket, struct query *ctl,
 	    {
 		if (suppress_readbody)
 		{
-		    /* When readheaders returns PS_TRUNCATED,
-		     * the body (which has no content)
-		     * has already been read by readheaders,
-		     * so we say readbody returned PS_SUCCESS
-		     */
 		    err = PS_SUCCESS;
 		}
 		else
@@ -618,7 +629,8 @@ static int fetch_messages(int mailserver_socket, struct query *ctl,
 		ctl->errcount++;
 		suppress_delete = TRUE;
 	    }
-	    (*fetches)++;
+	    if (!retained)
+		(*fetches)++;
 	}
 
 	/*
@@ -632,8 +644,14 @@ static int fetch_messages(int mailserver_socket, struct query *ctl,
 	 * now.
 	 */
 
-	/* tell the UID code we've seen this */
-	if (ctl->newsaved)
+	/*
+	 * Tell the UID code we've seen this.
+	 * Matthias Andree: only register the UID if we could actually
+	 * forward this mail. If we omit this !suppress_delete check,
+	 * fetchmail will never retry mail that the local listener
+	 * refused temporarily.
+	 */
+	if (ctl->newsaved && !suppress_delete)
 	{
 	    struct idlist	*sdp;
 
@@ -726,14 +744,13 @@ const int maxfetch;		/* maximum number of messages to fetch */
 	sigfillset(&allsigs);
 	sigprocmask(SIG_UNBLOCK, &allsigs, NULL);
 #endif /* HAVE_SIGPROCMASK */
-
+	
 	if (js == THROW_SIGPIPE)
 	{
 	    signal(SIGPIPE, SIG_IGN);
 	    report(stdout,
 		   GT_("SIGPIPE thrown from an MDA or a stream socket error\n"));
-	    err = PS_SOCKET;
-	    goto cleanUp;
+	    wait(0);
 	}
 	else if (js == THROW_TIMEOUT)
 	{
@@ -782,20 +799,10 @@ const int maxfetch;		/* maximum number of messages to fetch */
 		close_warning_by_mail(ctl, (struct msgblk *)NULL);
 		ctl->wedged = TRUE;
 	    }
-
-	    err = PS_ERROR;
 	}
 
-	/* try to clean up all streams */
-	release_sink(ctl);
-	if (ctl->smtp_socket != -1) {
-	    cleanupSockClose(ctl->smtp_socket);
-	    ctl->smtp_socket = -1;
-	}
-	if (mailserver_socket != -1) {
-	    cleanupSockClose(mailserver_socket);
-	    mailserver_socket = -1;
-	}
+	err = PS_SOCKET;
+	goto cleanUp;
     }
     else
     {
@@ -991,10 +998,14 @@ const int maxfetch;		/* maximum number of messages to fetch */
 	    phase = oldphase;
 	    goto closeUp;
 	}
-	set_timeout(0);
-	phase = oldphase;
 
 #ifdef SSL_ENABLE
+	/* Save the socket opened. Usefull if Fetchmail hangs on SSLOpen 
+	 * because the socket can be closed
+	 */
+	mailserver_socket_temp = mailserver_socket;
+	set_timeout(mytimeout);
+
 	/* perform initial SSL handshake on open connection */
 	/* Note:  We pass the realhost name over for certificate
 		verification.  We may want to make this configurable */
@@ -1004,10 +1015,20 @@ const int maxfetch;		/* maximum number of messages to fetch */
 	    report(stderr, GT_("SSL connection failed.\n"));
 	    goto closeUp;
 	}
+	
+	/* Fetchmail didn't hang on SSLOpen, 
+	 * then no need to set mailserver_socket_temp 
+	 */
+	mailserver_socket_temp = -1;
 #endif
-
+	
+	/* A timeout is still defined before SSLOpen, 
+	 * then Fetchmail hanging on SSLOpen is handled.
+	 */
+	set_timeout(0);
+	phase = oldphase;
 #ifdef KERBEROS_V4
-	if (ctl->server.authenticate == A_KERBEROS_V4)
+	if (ctl->server.authenticate == A_KERBEROS_V4 && (strcasecmp(proto->name,"IMAP") != 0))
 	{
 	    set_timeout(mytimeout);
 	    err = kerberos_auth(mailserver_socket, ctl->server.truename,
@@ -1110,6 +1131,12 @@ is restored."));
 			close_warning_by_mail(ctl, (struct msgblk *)NULL);
 		    }
 		}
+		else if (err == PS_REPOLL)
+		{
+		    report(stderr, GT_("Repoll immediately on %s@%s\n"),
+			   ctl->remotename,
+			   ctl->server.truename);
+		}
 		else
 		    report(stderr, GT_("Unknown login or authentication error on %s@%s\n"),
 			   ctl->remotename,
@@ -1285,11 +1312,26 @@ is restored."));
 		     */
 		    force_retrieval = !peek_capable && (ctl->errcount > 0);
 
+		    /*
+		     * Don't trust the message count passed by the server.
+		     * Without this check, it might be possible to do a
+		     * DNS-spoofing attack that would pass back a ridiculous 
+		     * count, and allocate a malloc area that would overlap
+		     * a portion of the stack.
+		     */
+		    if (count > INT_MAX/sizeof(int))
+		    {
+			report(stderr, GT_("bogus message count!"));
+			return(PS_PROTOCOL);
+		    }
+
 		    /* OK, we're going to gather size info next */
 		    xalloca(msgsizes, int *, sizeof(int) * count);
 		    xalloca(msgcodes, int *, sizeof(int) * count);
-		    for (i = 0; i < count; i++)
+		    for (i = 0; i < count; i++) {
+			msgsizes[i] = 0;
 			msgcodes[i] = MSGLEN_UNKNOWN;
+		    }
 
 		    /* 
 		     * We need the size of each message before it's
@@ -1322,6 +1364,8 @@ is restored."));
 			    continue;
 			else if (ctl->server.base_protocol->is_old && (ctl->server.base_protocol->is_old)(mailserver_socket,ctl,num))
 			    msgcodes[num-1] = MSGLEN_OLD;
+/*			else if (msgsizes[num-1] == 512)
+				msgcodes[num-1] = MSGLEN_OLD;  (hmh) sample code to skip message */
 		    }
 
 		    /* read, forward, and delete messages */
@@ -1365,12 +1409,26 @@ is restored."));
 
     cleanUp:
 	/* we only get here on error */
-	if (err != 0 && err != PS_SOCKET)
+	if (err != 0 && err != PS_SOCKET && err != PS_REPOLL)
 	{
 	    stage = STAGE_LOGOUT;
 	    (ctl->server.base_protocol->logout_cmd)(mailserver_socket, ctl);
 	}
-	cleanupSockClose(mailserver_socket);
+
+	/* try to clean up all streams */
+	release_sink(ctl);
+	smtp_close(ctl, 0);
+	if (mailserver_socket != -1) {
+	    cleanupSockClose(mailserver_socket);
+	    mailserver_socket = -1;
+	}
+	/* If there was a connect timeout, the socket should be closed.
+	 * mailserver_socket_temp contains the socket to close.
+	 */
+	if (mailserver_socket_temp != -1) {
+	    cleanupSockClose(mailserver_socket_temp);
+	    mailserver_socket_temp = -1;
+	}
     }
 
     msg = (const char *)NULL;	/* sacrifice to -Wall */
