@@ -11,6 +11,7 @@
 #include  <ctype.h>
 #if defined(STDC_HEADERS)
 #include  <stdlib.h>
+#include  <limits.h>
 #endif
 #include  "fetchmail.h"
 #include  "socket.h"
@@ -29,8 +30,9 @@ extern char *strstr();	/* needed on sysV68 R3V7.1. */
 #define IMAP4		0	/* IMAP4 rev 0, RFC1730 */
 #define IMAP4rev1	1	/* IMAP4 rev 1, RFC2060 */
 
-static int count, unseen, deletions, imap_version, preauth; 
-static int expunged, expunge_period, saved_timeout;
+static int count = 0, recentcount = 0, unseen = 0, deletions = 0;
+static int expunged, expunge_period, saved_timeout = 0;
+static int imap_version, preauth;
 static flag do_idle;
 static char capabilities[MSGBUFSIZE+1];
 static unsigned int *unseen_messages;
@@ -58,6 +60,19 @@ static int imap_ok(int sock, char *argbuf)
 	else if (strstr(buf, "EXISTS"))
 	{
 	    count = atoi(buf+2);
+	    /*
+	     * Don't trust the message count passed by the server.
+	     * Without this check, it might be possible to do a
+	     * DNS-spoofing attack that would pass back a ridiculous 
+	     * count, and allocate a malloc area that would overlap
+	     * a portion of the stack.
+	     */
+	    if (count > INT_MAX/sizeof(int))
+	    {
+		report(stderr, "bogus message count!");
+		return(PS_PROTOCOL);
+	    }
+
 	    /*
 	     * Nasty kluge to handle RFC2177 IDLE.  If we know we're idling
 	     * we can't wait for the tag matching the IDLE; we have to tell the
@@ -76,6 +91,10 @@ static int imap_ok(int sock, char *argbuf)
 		stage = STAGE_FETCH;
 	    }
 	}
+	else if (strstr(buf, "RECENT"))
+	{
+	    recentcount = atoi(buf+2);
+	}
 	else if (strstr(buf, "PREAUTH"))
 	    preauth = TRUE;
 	/*
@@ -181,7 +200,7 @@ static int do_imap_ntlm(int sock, struct query *ctl)
     if ((gen_recv(sock, msgbuf, sizeof msgbuf)))
 	return result;
   
-    len = from64tobits ((unsigned char*)&challenge, msgbuf, sizeof(msgbuf));
+    len = from64tobits ((char*)&challenge, msgbuf, sizeof(challenge));
     
     if (outlevel >= O_DEBUG)
 	dumpSmbNtlmAuthChallenge(stdout, &challenge);
@@ -337,6 +356,27 @@ static int imap_getauth(int sock, struct query *ctl, char *greeting)
     }
 #endif /* KERBEROS_V4 */
 
+#ifdef SSL_ENABLE
+    if ((ctl->server.authenticate == A_ANY)
+        && strstr(capabilities, "STARTTLS"))
+    {
+           char *realhost;
+
+           realhost = ctl->server.via ? ctl->server.via : ctl->server.pollname;
+           gen_transact(sock, "STARTTLS");
+
+           /* We use "tls1" instead of ctl->sslproto, as we want STARTTLS,
+            * not other SSL protocols
+            */
+           if (SSLOpen(sock,ctl->sslcert,ctl->sslkey,"tls1",ctl->sslcertck, ctl->sslcertpath,ctl->sslfingerprint,realhost,ctl->server.pollname) == -1)
+           {
+               report(stderr,
+                      GT_("SSL connection failed.\n"));
+               return(PS_AUTHFAIL);
+           }
+    }
+#endif /* SSL_ENABLE */
+
     /*
      * No such luck.  OK, now try the variants that mask your password
      * in a challenge-response.
@@ -420,7 +460,7 @@ static int imap_getauth(int sock, struct query *ctl, char *greeting)
 	imap_canonicalize(remotename, ctl->remotename, NAMELEN);
 	imap_canonicalize(password, ctl->password, PASSWORDLEN);
 
-	strcpy(shroud, ctl->password);
+	strcpy(shroud, password);
 	ok = gen_transact(sock, "LOGIN \"%s\" \"%s\"", remotename, password);
 	shroud[0] = '\0';
 	if (ok)
@@ -486,25 +526,41 @@ static int imap_getrange(int sock,
 	 * just after deletion.
 	 */
 	ok = 0;
-	if (deletions && expunge_period != 1)
+	if (deletions) {
 	    ok = internal_expunge(sock);
-	count = -1;
-	if (do_idle)
+	    if (ok)
+	    {
+		report(stderr, GT_("expunge failed\n"));
+		return(ok);
+	    }
+	}
+
+	/*
+	 * recentcount is already set here by the last imap command which
+	 * returned RECENT on detecting new mail. if recentcount is 0, wait
+	 * for new mail.
+	 */
+
+	/* this is a while loop because imap_idle() might return on other
+	 * mailbox changes also */
+	while (recentcount == 0 && do_idle) {
+	    smtp_close(ctl, 1);
 	    ok = imap_idle(sock);
-	if (ok || gen_transact(sock, "NOOP"))
-	{
-	    report(stderr, GT_("re-poll failed\n"));
-	    return(ok);
-	}
-	else if (count == -1)	/* no EXISTS response to NOOP/IDLE */
-	{
-	    count = 0;
+	    if (ok)
+	    {
+		report(stderr, GT_("re-poll failed\n"));
+		return(ok);
+	    }
 	}
+	/* if recentcount is 0, return no mail */
+	if (recentcount == 0)
+		count = 0;
 	if (outlevel >= O_DEBUG)
 	    report(stdout, GT_("%d messages waiting after re-poll\n"), count);
     }
     else
     {
+	count = 0;
 	ok = gen_transact(sock, 
 			  check_only ? "EXAMINE \"%s\"" : "SELECT \"%s\"",
 			  folder ? folder : "INBOX");
@@ -517,11 +573,35 @@ static int imap_getrange(int sock,
 	    report(stdout, GT_("%d messages waiting after first poll\n"), count);
 
 	/* no messages?  then we may need to idle until we get some */
-	if (count == 0 && do_idle)
-	    imap_idle(sock);
+	while (count == 0 && do_idle) {
+	    ok = imap_idle(sock);
+	    if (ok)
+	    {
+		report(stderr, GT_("re-poll failed\n"));
+		return(ok);
+	    }
+	}
+
+	/*
+	 * We should have an expunge here to
+	 * a) avoid fetching deleted mails during 'fetchall'
+	 * b) getting a wrong count of mails during 'no fetchall'
+	 */
+	if (!check_only && !ctl->keep && count > 0)
+	{
+	    ok = internal_expunge(sock);
+	    if (ok)
+	    {
+		report(stderr, GT_("expunge failed\n"));
+		return(ok);
+	    }
+	    if (outlevel >= O_DEBUG)
+		report(stdout, GT_("%d messages waiting after expunge\n"), count);
+	}
     }
 
     *countp = count;
+    recentcount = 0;
 
     /* OK, now get a count of unseen messages and their indices */
     if (!ctl->fetchall && count > 0)
@@ -575,7 +655,9 @@ static int imap_getrange(int sock,
 	unseen = -1;
 
     *newp = unseen;
+    count = 0;
     expunged = 0;
+    deletions = 0;
 
     return(PS_SUCCESS);
 }
@@ -620,7 +702,7 @@ static int imap_getsizes(int sock, int count, int *sizes)
      * known-bad size value.
      */
     if (count == 1)
-	gen_send(sock, "FETCH 1 RFC822.SIZE", count);
+	gen_send(sock, "FETCH 1 RFC822.SIZE");
     else
 	gen_send(sock, "FETCH 1:%d RFC822.SIZE", count);
     for (;;)
@@ -630,6 +712,9 @@ static int imap_getsizes(int sock, int count, int *sizes)
 
 	if ((ok = gen_recv(sock, buf, sizeof(buf))))
 	    return(ok);
+	/* an untagged NO means that a message was not readable */
+	else if (strstr(buf, "* NO"))
+	    ;
 	else if (strstr(buf, "OK") || strstr(buf, "NO"))
 	    break;
 	else if (sscanf(buf, "* %u FETCH (RFC822.SIZE %u)", &num, &size) == 2) {
@@ -702,9 +787,19 @@ static int imap_fetch_headers(int sock, struct query *ctl,int number,int *lenp)
   	    break;
 	/* try to recover from chronically fucked-up M$ Exchange servers */
  	else if (!strncmp(ptr, "NO", 2))
+	{
+	    /* wait for a tagged response */
+	    if (strstr (buf, "* NO"))
+		imap_ok (sock, 0);
  	    return(PS_TRANSIENT);
+	}
  	else if (!strncmp(ptr, "BAD", 3))
+	{
+	    /* wait for a tagged response */
+	    if (strstr (buf, "* BAD"))
+		imap_ok (sock, 0);
  	    return(PS_TRANSIENT);
+	}
     }
 
     if (num != number)
