@@ -20,6 +20,9 @@
 #ifdef HAVE_NET_SELECT_H /* AIX needs this */
 #include <net/select.h>
 #endif
+#ifdef HAVE_SYS_SELECT_H /* AIX 4.1, at least, needs this */
+#include  <sys/select.h>
+#endif
 #include  <netdb.h>
 #include  <errno.h>
 #include  <unistd.h>
