@@ -40,7 +40,7 @@ extern char *strstr();	/* needed on sysV68 R3V7.1. */
 
 int mytimeout;		/* value of nonreponse timeout */
 int suppress_tags;	/* emit tags? */
-char shroud[PASSWORDLEN];	/* string to shroud in debug output */
+char shroud[PASSWORDLEN*2+1];	/* string to shroud in debug output */
 struct msgblk msgblk;
 
 char tag[TAGLEN];
@@ -92,6 +92,13 @@ static void find_server_names(const char *hdr,
 	{
 	    char	*atsign;
 
+	    /* 
+	     * Handle empty address from a To: header containing only 
+	     * a comment.
+	     */
+	    if (!*cp)
+		continue;
+
 	    /*
 	     * If the name of the user begins with a qmail virtual
 	     * domain prefix, ignore the prefix.  Doing this here
@@ -474,6 +481,7 @@ int readheaders(int sock,
 	     */
 	    if (protocol->delimited && line[0] == '.' && EMPTYLINE(line+1))
 	    {
+		headers_ok = FALSE;
 		has_nuls = (linelen != strlen(line));
 		free(line);
 		goto process_headers;
@@ -1089,33 +1097,20 @@ int readheaders(int sock,
 #else
 		    sprintf(buf+1,
 #endif /* HAVE_SNPRINTF */
-			    "for %s@%s (by default); ",
-			    user, ctl->destaddr);
+			    "for %s (by default); ",
+			    rcpt_address (ctl, run.postmaster, 0));
 		}
 		else if (good_addresses == 1)
 		{
 		    for (idp = msgblk.recipients; idp; idp = idp->next)
 			if (idp->val.status.mark == XMIT_ACCEPT)
 			    break;	/* only report first address */
-		    if (strchr(idp->id, '@'))
 #ifdef HAVE_SNPRINTF
 		    snprintf(buf+1, sizeof(buf)-1,
 #else                       
 		    sprintf(buf+1,
 #endif /* HAVE_SNPRINTF */
-			    "for %s", idp->id);
-		    else
-			/*
-			 * This could be a bit misleading, as destaddr is
-			 * the forwarding host rather than the actual 
-			 * destination.  Most of the time they coincide.
-			 */
-#ifdef HAVE_SNPRINTF
-		    	snprintf(buf+1, sizeof(buf)-1,
-#else                       
-			sprintf(buf+1,
-#endif /* HAVE_SNPRINTF */
-				"for %s@%s", idp->id, ctl->destaddr);
+			    "for %s", rcpt_address (ctl, idp->id, 1));
 		    sprintf(buf+strlen(buf), " (%s); ",
 			    MULTIDROP(ctl) ? "multi-drop" : "single-drop");
 		}
@@ -1333,6 +1328,23 @@ void init_transact(const struct method *proto)
     protocol = (struct method *)proto;
 }
 
+static void enshroud(char *buf)
+/* shroud a password in the given buffer */
+{
+    char *cp;
+
+    if (shroud[0] && (cp = strstr(buf, shroud)))
+    {
+       char    *sp;
+
+       sp = cp + strlen(shroud);
+       *cp = '*';
+       while (*sp)
+           *cp++ = *sp++;
+       *cp = '\0';
+    }
+}
+
 #if defined(HAVE_STDARG_H)
 void gen_send(int sock, const char *fmt, ... )
 #else
@@ -1372,18 +1384,7 @@ va_dcl
 
     if (outlevel >= O_MONITOR)
     {
-	char *cp;
-
-	if (shroud[0] && (cp = strstr(buf, shroud)))
-	{
-	    char	*sp;
-
-	    sp = cp + strlen(shroud);
-	    *cp++ = '*';
-	    while (*sp)
-		*cp++ = *sp++;
-	    *cp = '\0';
-	}
+	enshroud(buf);
 	buf[strlen(buf)-2] = '\0';
 	report(stdout, "%s> %s\n", protocol->name, buf);
     }
@@ -1462,19 +1463,8 @@ va_dcl
 
     if (outlevel >= O_MONITOR)
     {
-	char *cp;
-
-	if (shroud && shroud[0] && (cp = strstr(buf, shroud)))
-	{
-	    char	*sp;
-
-	    sp = cp + strlen(shroud);
-	    *cp++ = '*';
-	    while (*sp)
-		*cp++ = *sp++;
-	    *cp = '\0';
-	}
-	buf[strlen(buf)-1] = '\0';
+	enshroud(buf);
+	buf[strlen(buf)-2] = '\0';
 	report(stdout, "%s> %s\n", protocol->name, buf);
     }
 
